/* ctime.c - a Ruby interface to Corporate Time
 * 
 * $Id: ctime.c,v 1.49 2004/11/07 07:36:01 ianmacd Exp $
 *
 * Author: Ian Macdonald <ian@caliban.org>
 *
 * Copyright (C) 2002-2004 Ian Macdonald
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ruby.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* Required for ctapi.h */
#ifdef WIN32
#define CAPI_FUNC(a)  __declspec(dllimport) a
#else
#define CAPI_FUNC(a) a
#endif

#include <ctapi.h>
#include "capilayer.h"

#define CTIME_VERSION   "0.3.2"
#define CTIME_STRFTIME	"%Y%m%dT%H%M00Z"
#define MAX_HANDLES	32

/* template for an MIME wrapper icalendar event */
static const char *MIME_HEADER = {
    "MIME-Version: 1.0\n"
	"Content-Type: text/calendar\n"
	"Content-Transfer-Encoding: quoted-printable\n\n" "%s"
};

static const char *ICAL_TEMPLATE = {
    "BEGIN:VCALENDAR\n"
	"VERSION:2.0\n"
	"PRODID://Steltor//CAPI//EN\n"
	"BEGIN:VEVENT\n"
	"CLASS:PUBLIC\n" "%s" "END:VEVENT\n" "END:VCALENDAR\n"
};

int agenda_count = 0;
const char *error_name = NULL;
CAPISession session = NULL;
CAPIStatus stat = 0;
CAPIStream my_stream = NULL;

/*
 * Leaving an extra slot for null terminating the array.
 */
CAPIHandle agenda[MAX_HANDLES + 1];

/* name of this class */
VALUE cCTime;

/* classes referred to */
VALUE cEvent;
VALUE cAgenda;

/* these are exceptions in the CTime class */
VALUE eCTime_AgendaError;
VALUE eCTime_EventError;
VALUE eCTime_DesignationError;
VALUE eCTime_UnknownUserError;

/* CTime.initialize */
static VALUE ct_init(VALUE self, VALUE server, VALUE user, VALUE passwd)
{
    const char *txt = NULL;
    const char *comma = ",";
    const char *comma_sp = ", ";
    VALUE capabilities;
    VALUE cap_list;
    VALUE rv;

    rb_ctime_debug1("Connecting to server %s...\n", STR2CSTR(server));
    stat = capi_connect(STR2CSTR(server), STR2CSTR(user), STR2CSTR(passwd),
                        &session);

    capi_get_status_string(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);


    if (stat == CAPI_STAT_OK) {

	/* list of e-mail addresses pertaining to open agendas */
	rb_iv_set(self, "@agendas", rb_ary_new());

	/* retrieve information on capabilities */
	capabilities = rb_hash_new();

        capi_get_capabilities(session, CAPI_CAPAB_AUTH, &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma);
	rb_hash_aset(capabilities, rb_str_new2("authentication"),
		     cap_list);

        capi_get_capabilities(session, CAPI_CAPAB_COMP, &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma);
	rb_hash_aset(capabilities, rb_str_new2("compression"), cap_list);

        capi_get_capabilities(session, CAPI_CAPAB_ENCR, &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma);
	rb_hash_aset(capabilities, rb_str_new2("encryption"), cap_list);

        capi_get_capabilities(session, CAPI_CAPAB_MAXDATE, &txt);
	rb_hash_aset(capabilities, rb_str_new2("max_date"),
		     rb_str_new2(txt));

        capi_get_capabilities(session, CAPI_CAPAB_CAPI_VERSION, &txt);
	rb_hash_aset(capabilities, rb_str_new2("capi_version"),
		     rb_str_new2(txt));

        capi_get_capabilities(session, CAPI_CAPAB_VERSION, &txt);
	rb_hash_aset(capabilities, rb_str_new2("version"),
		     rb_str_new2(txt));

        capi_get_capabilities(session, CAPI_CAPAB_SERVER_VERSION, &txt);
	rb_hash_aset(capabilities, rb_str_new2("server_version"),
		     rb_str_new2(txt));

        capi_get_capabilities(session, CAPI_CAPAB_UNSUPPORTED_ICAL_PROP, &txt);
	rv = ruby_verbose;
	ruby_verbose = Qnil; /* work around warning in 1.8.8 */
	cap_list = rb_str_split(rb_str_new2(txt), comma_sp);
	ruby_verbose = rv;
	rb_hash_aset(capabilities, rb_str_new2("unsupported_ical_prop"),
		     cap_list);

        capi_get_capabilities(session, CAPI_CAPAB_UNSUPPORTED_ICAL_COMP, &txt);
	ruby_verbose = Qnil; /* work around warning in 1.8.8 */
	cap_list = rb_str_split(rb_str_new2(txt), comma_sp);
	ruby_verbose = rv;
	rb_hash_aset(capabilities, rb_str_new2("unsupported_ical_comp"),
		     cap_list);

	rb_iv_set(self, "@capabilities", capabilities);

	return self;

    } else {

	return Qnil;		/* but why don't we actually get nil
				   assigned to our variable when object
				   instantiation fails? Maybe it's because
				   we return control to CTime.new */
    }
}

/* reader function for @error */
static VALUE ct_error(VALUE self)
{
    return rb_iv_get(self, "@error");
}

/* reader function for @agendas */
static VALUE ct_agendas(VALUE self)
{
    return rb_iv_get(self, "@agendas");
}

/* reader function for @capabilities */
static VALUE ct_capabilities(VALUE self)
{
    return rb_iv_get(self, "@capabilities");
}

/* close opened agendas and disconnect from the server */
static VALUE ct_quit(VALUE self)
{
    rb_ctime_debug0("Quitting from server...\n");

    if (session) {
	if (agenda_count > 0) {
	    rb_ctime_debug1("There are %d agendas open\n", agenda_count);

	    rb_ctime_debug0("Destroying handles...\n");
            stat = capi_destroy_multiple_handles(session, agenda,
                                                 agenda_count);
            capi_get_status_string(stat, &error_name);
	    agenda_count = 0;

	    /* set @error */
	    rb_iv_set(self, "@error", rb_str_new2(error_name));
	    rb_ctime_debug1("Status = %s\n", error_name);
	}

	rb_ctime_debug0("Logging off...\n");
        stat = capi_disconnect(&session);
        capi_get_status_string(stat, &error_name);

	/* set @error */
	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);

	return self;
    }

    return Qnil;
}

/* open a new agenda */
static VALUE ct_open_agenda(VALUE self, VALUE user)
{
    VALUE agendas;
    VALUE new_agenda;
    const char *info = NULL;

    if (agenda_count >= MAX_HANDLES) {
        rb_raise(eCTime_AgendaError, "Too many agendas open");
        return Qnil;
    }

    rb_ctime_debug1("Opening agenda for %s...\n", STR2CSTR(user));
    agenda[agenda_count] = NULL;

    rb_ctime_debug0("Getting handle...\n");
    stat = capi_get_handle(session, STR2CSTR(user), &agenda[agenda_count]);

    capi_get_status_string(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat == CAPI_STAT_OK) {

        rb_ctime_debug0("Creating agenda object...\n");
        new_agenda = Data_Wrap_Struct(cAgenda, 0, 0, &agenda[agenda_count]);
        rb_ctime_debug0("Calling init...\n");
        rb_obj_call_init(new_agenda, 0, 0);
        rb_ctime_debug0("Pushing agenda onto @agendas...\n");

        agendas = rb_iv_get(self, "@agendas");
        rb_ary_push(agendas, new_agenda);

        rb_ctime_debug0("Querying agenda info...\n");

        rb_ctime_debug0("Getting handle info...\n");
        stat = capi_get_handle_info(session, agenda[agenda_count],
                                    CAPI_HANDLE_TYPE, &info);
        if (stat == CAPI_STAT_OK) {
            rb_ctime_debug0("Got type...\n");
            rb_iv_set(new_agenda, "@type", rb_str_new2(info));
        }

        stat = capi_get_handle_info(session, agenda[agenda_count],
                                    CAPI_HANDLE_NAME, &info);
        if (stat == CAPI_STAT_OK) {
            rb_ctime_debug0("Got name...\n");
            rb_iv_set(new_agenda, "@name", rb_str_new2(info));
        }

        stat = capi_get_handle_info(session, agenda[agenda_count],
                                    CAPI_HANDLE_MAILTO, &info);
        if (stat == CAPI_STAT_OK) {
            rb_ctime_debug0("Got mailto...\n");
            rb_iv_set(new_agenda, "@mailto", rb_str_new2(info));
        }

        ++agenda_count;
        /*
         * Null terminate the array for ct_get_events
         */
        agenda[agenda_count] = NULL;
	rb_ctime_debug1("There are %d open agendas\n", agenda_count);
        return new_agenda;

    } else if (stat & CAPI_STAT_DATA_USERID) {

	/* no user on the server matched this name */
	rb_raise(eCTime_AgendaError,
		 "Agenda for %s does not exist or password incorrect.",
		 STR2CSTR(user));
    }

    /* shouldn't get here very often */
    rb_ctime_debug0("Program should not reach here\n");

    return Qnil;

}

/* close an open agenda */
static VALUE ct_close_agenda(VALUE self, VALUE close_agenda)
{
    CAPIHandle *agenda_handle;
    VALUE agendas;

    Data_Get_Struct(close_agenda, CAPIHandle, agenda_handle);

    rb_ctime_debug1("Closing agenda for %s...\n", 
            STR2CSTR(rb_iv_get(close_agenda, "@name")));

    if (agenda_count) {
	int found = 0;
	int i, j;

	for (i = 1; i <= agenda_count; ++i) {

            if (agenda_handle == &agenda[i-1]) {

		found = 1;
                rb_ctime_debug0("Found matching agenda...\n");
                stat = capi_destroy_multiple_handles(session, &agenda[i - 1],
                                                     1);
                rb_ctime_debug0("Finished destroying agenda...\n");
                capi_get_status_string(stat, &error_name);
		rb_iv_set(self, "@error", rb_str_new2(error_name));
		rb_ctime_debug1("Status = %s\n", error_name);

		agendas = rb_iv_get(self, "@agendas");

		/* close up hole in array */
		for (j = i - 1; j < agenda_count; ++j) {
		    agenda[j] = agenda[j + 1];
		}

		/* remove agenda from @agendas */
		for (j = 1; j <= agenda_count; ++j) {
                    if (close_agenda == rb_ary_entry(agendas, j - 1)) {
			rb_ary_delete_at(agendas, j - 1);
			break;
		    }
		}

		--agenda_count;
	    }
	}

	if (!found)
	    rb_raise(eCTime_AgendaError, "No open agenda for %s",
                     STR2CSTR(rb_iv_get(close_agenda, "@name")));

        rb_ctime_debug1("Closed agenda for %s.\n",
                        STR2CSTR(rb_iv_get(close_agenda, "@name")));
        return close_agenda;

    } else {
	rb_raise(eCTime_AgendaError, "No agendas opened");
    }
}

/* get the events for a given agenda */
static VALUE ct_get_events(int argc, VALUE *argv, VALUE self)
{
    const char *iCalData = NULL;
    char *strf_fmt = "%Y%m%dT%H%M00Z";
    VALUE start_time;
    VALUE end_time;
    VALUE should_parse;
    VALUE event_list;

    rb_scan_args(argc, argv, "21", &start_time, &end_time, &should_parse);
    if (should_parse == Qnil)
      should_parse = Qtrue;

    rb_ctime_debug0("Preparing to get events...\n");

    if (agenda_count == 0)
	rb_raise(eCTime_AgendaError, "No agendas opened");

    /* Check whether start_time or end_time responds to gmtime. If so,
     * convert values to GMT. This will not affect an object that has
     * already been converted to GMT at the Ruby level.
     */
    if (rb_respond_to(start_time, rb_intern("gmtime")) != 0) {
      start_time = rb_funcall(start_time, rb_intern("gmtime"), 0);
    } else if (rb_respond_to(start_time, rb_intern("new_offset")) != 0) {
      start_time = rb_funcall(start_time,
			      rb_intern("new_offset"), 1, INT2NUM(0));
    }

    if (rb_respond_to(end_time, rb_intern("gmtime")) != 0) {
      end_time = rb_funcall(end_time, rb_intern("gmtime"), 0);
    } else if (rb_respond_to(end_time, rb_intern("new_offset")) != 0) {
      end_time = rb_funcall(end_time,
			    rb_intern("new_offset"), 1, INT2NUM(0));
    }

    /* Check whether start_time or end_time responds to strftime. If so,
     * convert it to a String.
     */
    if (rb_respond_to(start_time, rb_intern("strftime")) != 0) {
      start_time = rb_funcall(start_time, rb_intern("strftime"),
			      1, rb_str_new2(strf_fmt));
    }
    if (rb_respond_to(end_time, rb_intern("strftime")) != 0) {
      end_time = rb_funcall(end_time, rb_intern("strftime"),
			      1, rb_str_new2(strf_fmt));
    }

    rb_ctime_debug2("Will fetch events in range %s to %s...\n",
		    STR2CSTR(start_time), STR2CSTR(end_time));

    /* create a memory stream to receive the information */
    rb_ctime_debug0("Creating memory stream...\n");
    stat = capi_create_memory_stream(session, &my_stream, NULL, &iCalData);

    /*
    iCalData = rb_str_new2("");
    stat = capi_create_read_callback_stream(session, &my_stream, iCalData);
    */

    capi_get_status_string(stat, &error_name);
	    rb_iv_set(self, "@error", rb_str_new2(error_name));
	    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat != CAPI_STAT_OK)
        return Qnil;

    rb_ctime_debug0("Fetching all properties...\n");
    stat = capi_fetch_events_by_range(session, agenda, agenda_count,
                                      STR2CSTR(start_time),
				      STR2CSTR(end_time), NULL, 0, my_stream);

    capi_get_status_string(stat, &error_name);
	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);

    /*
     * The api returns CAPI_STAT_SECUR_READ_PROPS to indicate that some of the
     * requested properties weren't returned (i.e. there are private meetings).
     * It's okay though - the stream is still populated with only the readable
     * attributes for the private meetings and all the attributes for public
     * ones.
     */
    if ((stat != CAPI_STAT_OK) && (stat != CAPI_STAT_SECUR_READ_PROPS)) {
	/* release the stream */
	rb_ctime_debug0("Destroying streams...\n");
        capi_destroy_stream(session, &my_stream);

	return Qnil;
    }

    if (should_parse == Qtrue) {
	rb_ctime_debug0("Calling parse_events...\n");
	event_list = rb_funcall(cEvent, rb_intern("parse_events"),
				1, rb_str_new2(iCalData));
    } else {
	rb_ctime_debug0("Returning raw iCal stream...\n");
	event_list = rb_str_new2(iCalData);
    }

    /* Release the stream */
    rb_ctime_debug0("Destroying streams...\n");

    stat = capi_destroy_stream(session, &my_stream);

    capi_get_status_string(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    rb_ctime_debug0("get_events() finished.\n");

    return event_list;
}

/* work as a designate */
static VALUE ct_designate(VALUE self, VALUE user)
{
    rb_ctime_debug1("Setting identity to %s\n", user);
    stat = capi_set_identity(session, STR2CSTR(user));

    if (stat & CAPI_STAT_SECUR_LOGON_AUTH) {
	/* This user has not granted you designate rights! */
	rb_raise(eCTime_AgendaError,
		 "Insufficient designation rights for user %s",
		 STR2CSTR(user));
    } else if (stat & CAPI_STAT_DATA_USERID) {
	/* No user on the server matched this name! */
	rb_raise(eCTime_UnknownUserError, "User %s unknown",
		 STR2CSTR(user));
    }

    capi_get_status_string(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat != CAPI_STAT_OK)
	return Qnil;

    return Qtrue;
}

/* create an event */
static VALUE ct_set_event(VALUE self, VALUE event)
{
    char ical_event[1024];
    char new_event[65536];
    char tmp_buffer[32768];
    char uid_buffer[128];
    const char *new_uid = NULL;
    int error = 0;

    rb_ctime_debug0("Creating a new event on the server...");

    if (agenda_count == 0)
	rb_raise(eCTime_AgendaError, "No agendas opened");

    sprintf(new_event, "DTSTART:%s\n",
            STR2CSTR(rb_funcall(event, rb_intern("start_time_str"), 0)));

    if (rb_iv_get(event, "@uid") != Qnil) {
	sprintf(tmp_buffer, "UID:%s\n",
		STR2CSTR(rb_iv_get(event, "@uid")));
	strcat(new_event, tmp_buffer);

	sprintf(tmp_buffer, "RECURRENCE-ID:%s\n",
                STR2CSTR(rb_funcall(event, rb_intern("start_time_str"), 0)));
	strcat(new_event, tmp_buffer);
    }

    sprintf(tmp_buffer, "DURATION:%s\n",
	    STR2CSTR(rb_funcall(event, rb_intern("duration_str"), 0)));
    strcat(new_event, tmp_buffer);

    if (rb_iv_get(event, "@summary") != Qnil) {
	sprintf(tmp_buffer, "SUMMARY:%s\n",
		STR2CSTR(rb_iv_get(event, "@summary")));
	strcat(new_event, tmp_buffer);
    }

    if (rb_iv_get(event, "@location") != Qnil) {
	sprintf(tmp_buffer, "LOCATION:%s\n",
		STR2CSTR(rb_iv_get(event, "@location")));
	strcat(new_event, tmp_buffer);
    }

    if (rb_iv_get(event, "@description") != Qnil) {
	sprintf(tmp_buffer, "DESCRIPTION:%s\n",
		STR2CSTR(rb_iv_get(event, "@description")));
	strcat(new_event, tmp_buffer);
    }

    sprintf(tmp_buffer, ICAL_TEMPLATE, new_event);
    sprintf(ical_event, MIME_HEADER, tmp_buffer);

    /* store the event */
    rb_ctime_debug0("Creating memory stream...\n");
    stat = capi_create_memory_stream(session, &my_stream, ical_event, NULL);
    capi_get_status_string(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat == CAPI_STAT_OK) {
	rb_ctime_debug0("Storing event...\n");
        stat = capi_store_event(session, agenda, agenda_count,
                                my_stream, &new_uid);

        capi_get_status_string(stat, &error_name);
	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);
    }

    if (stat != CAPI_STAT_OK) {
      rb_ctime_debug0("Destroying streams...\n");
      stat = capi_destroy_stream(session, &my_stream);
      capi_get_status_string(stat, &error_name);

	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);

      rb_raise(eCTime_AgendaError, "%s", error_name);

      return Qnil;
    }

    strcpy(uid_buffer, new_uid);
    rb_ctime_debug1("Got UID %s ...\n", uid_buffer);

    rb_ctime_debug0("Destroying streams...\n");
    stat = capi_destroy_stream(session, &my_stream);
    capi_get_status_string(stat, &error_name);

    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (error) {
	rb_raise(eCTime_AgendaError, "%s", error_name);

	return Qnil;

    }

    return rb_str_new2(uid_buffer);
}

/* delete an event */
static VALUE ct_delete_event(VALUE self, VALUE uid)
{
    rb_ctime_debug0("Deleting event from server...\n");

    if (agenda_count == 0)
	rb_raise(eCTime_AgendaError, "No agendas opened");

    stat = capi_delete_event(session, agenda, agenda_count, STR2CSTR(uid));
    capi_get_status_string(stat, &error_name);

    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat == CAPI_STAT_DATA_UID_NOTFOUND)
	rb_raise(eCTime_EventError, "No matching event for UID %s",
		 STR2CSTR(uid));

    if (stat != CAPI_STAT_OK)
	return Qnil;

    return Qtrue;
}

/* initialize this class */
void Init_ctime()
{
    rb_ctime_debug0("Inside Init_ctime...\n");

    /* define the CTime class */
    cCTime = rb_define_class("CTime", rb_cObject);

    /* define the CTime::AgendaError exception */
    eCTime_AgendaError = rb_define_class_under(cCTime, "AgendaError",
					       rb_eRuntimeError);

    /* define the CTime::EventError exception */
    eCTime_EventError = rb_define_class_under(cCTime, "EventError",
					      rb_eRuntimeError);

    /* define the CTime::EventError exception */
    eCTime_DesignationError =
	rb_define_class_under(cCTime, "DesignationError",
			      rb_eRuntimeError);

    /* define the CTime::UnknownUser exception */
    eCTime_UnknownUserError =
	rb_define_class_under(cCTime, "UnknownUserError",
			      rb_eRuntimeError);

    /* define CTime::VERSION constant */
    rb_define_const(cCTime, "VERSION", rb_str_new2(CTIME_VERSION));

    /* define CTime::STRFTIME constant */
    rb_define_const(cCTime, "STRFTIME", rb_str_new2(CTIME_STRFTIME));

    /* define the CTime.initialize method */
    rb_define_method(cCTime, "initialize", ct_init, 3);

    /* define some accessor methods */
    rb_define_method(cCTime, "error", ct_error, 0);
    rb_define_method(cCTime, "agendas", ct_agendas, 0);
    rb_define_method(cCTime, "capabilities", ct_capabilities, 0);

    /* define the CTime#quit method */
    rb_define_method(cCTime, "quit", ct_quit, 0);

    /* define the CTime#open_agenda method */
    rb_define_method(cCTime, "open_agenda", ct_open_agenda, 1);

    /* define the CTime#close_agenda method */
    rb_define_method(cCTime, "close_agenda", ct_close_agenda, 1);

    /* define the CTime#get_events method */
    rb_define_method(cCTime, "get_events", ct_get_events, -1);

    /* define the CTime#designate method */
    rb_define_method(cCTime, "designate", ct_designate, 1);

    /* define the CTime#set_event method */
    rb_define_method(cCTime, "set_event", ct_set_event, 1);

    /* define the CTime#delete_event method */
    rb_define_method(cCTime, "delete_event", ct_delete_event, 1);

    rb_ctime_debug0("Requiring event...\n");
    rb_require("ctime/event");
    rb_ctime_debug0("Requiring agenda...\n");
    rb_require("ctime/agenda");
    rb_ctime_debug0("Requiring attendee...\n");
    rb_require("ctime/attendee");

    rb_ctime_debug0("Grabbing Event and Agenda symbols...\n");
    cEvent = rb_eval_string("Event");
    cAgenda = rb_eval_string("Agenda");

    return;
}
